import os
import mathutils

class ObjectExporter:
	def __init__(self):
		self.show_progress = True
		self.use_strips = True
		self.use_degen_tris = False
		self.use_textures = True
		self.export_all = False
		self.collection = False
		self.shared_resources = True
		self.export_lods = True

	def compute_bounding_sphere(self, obj):
		p1 = max(((v.co, v.co.length) for v in obj.data.vertices), key=lambda x:x[1])[0]
		p2 = max(((v.co, (v.co-p1).length) for v in obj.data.vertices), key=lambda x:x[1])[0]
		center = (p1+p2)/2
		radius = (p1-p2).length/2
		for v in obj.data.vertices:
			d = v.co-center
			if d.length>radius:
				center += d*(1-radius/d.length)/2
				radius = (radius+d.length)/2

		return center, radius

	def collect_object_lods(self, obj):
		lods = [obj]
		if self.export_lods:
			lods += sorted([c for c in obj.children if c.lod_for_parent], key=(lambda l: l.lod_index))
			for i, l in enumerate(lods):
				if i>0 and l.lod_index!=i:
					raise Exception("Inconsistent LOD indices")

		return lods

	def create_mesh_exporter(self):
		from .export_mesh import MeshExporter
		mesh_export = MeshExporter()
		mesh_export.use_strips = self.use_strips
		mesh_export.use_degen_tris = self.use_degen_tris
		return mesh_export

	def create_material_exporter(self):
		from .export_material import MaterialExporter
		material_export = MaterialExporter()
		material_export.use_textures = self.use_textures
		return material_export

	def create_material_atlas_exporter(self):
		from .export_material import MaterialAtlasExporter
		material_atlas_export = MaterialAtlasExporter()
		return material_atlas_export

	def export_to_file(self, context, out_fn):
		if self.export_all:
			objs = [o for o in context.selected_objects if o.type=="MESH"]
		else:
			objs = [context.active_object]

		from .util import Progress
		progress = Progress(self.show_progress and context)

		path, base = os.path.split(out_fn)
		base, ext = os.path.splitext(base)

		resources = {}
		for i, obj in enumerate(objs):
			if self.export_all:
				out_fn = os.path.join(path, obj.name+ext)

			progress.push_task_slice(obj.name, i, len(objs))
			self.export_object_resources(context, obj, resources, progress)

			obj_res = self.export_object(context, obj, progress, resources=resources)
			refs = obj_res.collect_references()
			if not self.shared_resources:
				numbers = {}
				for r in refs:
					res_ext = os.path.splitext(r.name)[1]
					n = numbers.get(res_ext, 0)
					if n>0:
						r.name = "{}_{}{}".format(base, n, res_ext)
					else:
						r.name = base+res_ext
					numbers[res_ext] = n+1

			if self.collection:
				obj_res.write_collection(out_fn)
			else:
				for r in refs:
					r.write_to_file(os.path.join(path, r.name))
				obj_res.write_to_file(out_fn)

			progress.pop_task()

	def export_object_resources(self, context, obj, resources, progress, material_atlass=None):
		if material_atlass is None:
			material_atlass = {}

		lods = self.collect_object_lods(obj)

		from .mesh import create_mesh_from_object
		from .material import create_material_atlas
		mesh_export = self.create_mesh_exporter()
		material_export = self.create_material_exporter()
		material_atlas_export = self.create_material_atlas_exporter()

		for i, l in enumerate(lods):
			lod_index = l.lod_index if l.lod_for_parent else 0
			progress.push_task_slice("LOD {}".format(lod_index), i, len(lods))

			material_atlas = None
			atlas_flags = [m.render_mode!='EXTERNAL' and m.material_atlas for m in l.data.materials if m]
			if any(atlas_flags):
				mmk = lambda m: m.shader if m.render_mode=='CUSTOM' else ""
				material_atlas_key = mmk(l.data.materials[0])
				key_mismatch = any(mmk(m)!=material_atlas_key for m in l.data.materials)
				if not all(atlas_flags) or key_mismatch:
					raise Exception("Conflicting settings in object materials")

				if material_atlas_key in material_atlass:
					material_atlas = material_atlass[material_atlas_key]
				else:
					material_atlas = create_material_atlas(context, l.data.materials[0])
					material_atlass[material_atlas_key] = material_atlas

				tech_name = "{}.tech".format(material_atlas.name)
				if tech_name not in resources:
					material_atlas_export.export_technique_resources(material_atlas, resources)
					resources[tech_name] = material_atlas_export.export_technique(material_atlas, resources=resources)
			elif l.material_slots and l.material_slots[0].material:
				material = l.material_slots[0].material
				if material.render_mode!='EXTERNAL':
					tech_name = material.name+".tech"
					if tech_name not in resources:
						material_export.export_technique_resources(material, resources)
						resources[tech_name] = material_export.export_technique(material, resources=resources)
			elif "stub.tech" not in resources:
				resources["stub.tech"] = self.export_stub_technique()

			mesh_name = l.data.name+".mesh"
			if mesh_name not in resources:
				mesh = create_mesh_from_object(context, l, progress, material_atlas=material_atlas)
				mesh_res = mesh_export.export_mesh(context, mesh, progress)
				resources[mesh_name] = mesh_res

			progress.pop_task()

	def export_object(self, context, obj, progress, *, resources=None):
		if resources is None:
			resources = {}
			self.export_object_resources(context, obj, resources, progress)

		lods = self.collect_object_lods(obj)

		from .datafile import Resource, Statement
		obj_res = Resource(obj.name+".object", "object")
		statements = obj_res.statements

		center, radius = self.compute_bounding_sphere(obj)
		statements.append(Statement("bounding_sphere_hint", *center, radius))

		prev_mesh = None
		prev_tech = None
		for i, l in enumerate(lods):
			lod_st = []

			if l.data.name!=prev_mesh:
				mesh_res = resources[l.data.name+".mesh"]
				lod_st.append(obj_res.create_reference_statement("mesh", mesh_res))

				prev_mesh = l.data.name

			material = None
			if l.material_slots:
				material = l.material_slots[0].material
			if material:
				if material.render_mode=='EXTERNAL':
					tech_name = material.technique
				elif material.material_atlas:
					tech_name = "material_atlas_{}.tech".format(os.path.splitext(material.technique)[0])
				else:
					tech_name = material.name+".tech"
			else:
				tech_name = "stub.tech"

			if tech_name!=prev_tech:
				if material and material.render_mode=='EXTERNAL':
					lod_st.append(Statement("technique", material.technique))
				else:
					lod_st.append(obj_res.create_reference_statement("technique", resources[tech_name]))
				prev_tech = tech_name

			if i>0:
				st = Statement("level_of_detail", i)
				st.sub = lod_st
				statements.append(st)
			else:
				statements += lod_st

		progress.set_progress(1.0)

		return obj_res

	def export_stub_technique(self):
		from .datafile import Resource, Statement
		tech_res = Resource("stub.tech", "technique")
		pass_st = Statement("pass", "")
		tech_res.statements.append(pass_st)
		mat_st = Statement("material")
		pass_st.sub.append(mat_st)
		mat_st.sub.append(Statement("basic"))
		return tech_res
