#include <algorithm>
#include "ambientocclusion.h"
#include "blend.h"
#include "camera.h"
#include "renderer.h"
#include "resources.h"
#include "shader.h"
#include "tests.h"

using namespace std;

namespace Msp {
namespace GL {

AmbientOcclusion::AmbientOcclusion(Resources &resources, unsigned w, unsigned h, float):
	occlude_target(w, h, (RENDER_COLOR,R8)),
	occlude_shader(resources.get<Program>("_ambientocclusion_occlude.glsl.shader")),
	combine_shader(resources.get<Program>("_ambientocclusion_combine.glsl.shader")),
	quad(resources.get<Mesh>("_fullscreen_quad.mesh")),
	linear_sampler(resources.get<Sampler>("_linear_clamp.samp")),
	nearest_sampler(resources.get<Sampler>("_nearest_clamp.samp"))
{
	unsigned seed = 1;
	rotate_lookup.storage(RGBA8, 4, 4, 1);
	unsigned char data[64];
	for(unsigned i=0; i<16; ++i)
	{
		Geometry::Angle<float> a = Geometry::Angle<float>::from_turns(random(seed));
		unsigned char c = (cos(a)*0.5f+0.5f)*255;
		unsigned char s = (sin(a)*0.5f+0.5f)*255;
		data[i*4  ] = c;
		data[i*4+1] = s;
		data[i*4+2] = 255-s;
		data[i*4+3] = ((i+i/4)%2)*255;
	}
	rotate_lookup.image(0, data);

	shdata.uniform("inverse_projection", Matrix());

	set_n_samples(16);
	set_occlusion_radius(0.5f);
	set_darkness(1.0f);
	set_edge_depth_threshold(0.1f);
}

float AmbientOcclusion::random(unsigned &seed)
{
	static const unsigned modulus = (1U<<31)-1;
	seed = (static_cast<UInt64>(seed)*48271)%modulus;  // minstd
	return static_cast<float>(seed)/(modulus-1);
}

void AmbientOcclusion::set_n_samples(unsigned n)
{
	if(n<1 || n>32)
		throw out_of_range("AmbientOcclusion::set_n_samples");

	unsigned seed = 1;
	float radius_divisor = (n-1)*(n-1);
	Vector3 sample_points[32];
	for(unsigned i=0; i<n; ++i)
	{
		Vector3 v(random(seed)-0.5f, random(seed)-0.5f, random(seed)-0.5f);
		sample_points[i] = normalize(v)*(0.1f+0.9f*i*i/radius_divisor);
	}
	shdata.uniform3_array("sample_points", n, &sample_points[0].x);
	shdata.uniform("n_samples", static_cast<int>(n));
}

void AmbientOcclusion::set_occlusion_radius(float r)
{
	shdata.uniform("occlusion_radius", r);
}

void AmbientOcclusion::set_darkness(float darkness)
{
	shdata.uniform("darkness", darkness);
}

void AmbientOcclusion::set_edge_depth_threshold(float edt)
{
	shdata.uniform("edge_depth_threshold", edt);
}

void AmbientOcclusion::render(Renderer &renderer, const Texture2D &color, const Texture2D &depth)
{
	if(renderer.get_camera())
		shdata.uniform("inverse_projection", invert(renderer.get_camera()->get_projection_matrix()));

	Renderer::Push push(renderer);
	renderer.set_texture("source", &color, &nearest_sampler);
	renderer.set_texture("depth", &depth, &nearest_sampler);
	renderer.set_texture("occlusion", &occlude_target.get_target_texture(RENDER_COLOR), &linear_sampler);
	renderer.set_texture("rotate", &rotate_lookup, &nearest_sampler);
	renderer.set_shader_program(&occlude_shader, &shdata);

	{
		BindRestore bind_fbo(occlude_target.get_framebuffer());
		quad.draw(renderer);
	}

	renderer.set_shader_program(&combine_shader);
	quad.draw(renderer);
}


AmbientOcclusion::Template::Template():
	n_samples(16),
	occlusion_radius(0.5f),
	darkness(1.0f),
	edge_depth_threshold(0.1f)
{ }

AmbientOcclusion *AmbientOcclusion::Template::create(Resources &res, unsigned width, unsigned height) const
{
	RefPtr<AmbientOcclusion> ao = new AmbientOcclusion(res, width/size_divisor, height/size_divisor);
	ao->set_n_samples(n_samples);
	ao->set_occlusion_radius(occlusion_radius);
	ao->set_darkness(darkness);
	ao->set_edge_depth_threshold(edge_depth_threshold);
	return ao.release();
}


AmbientOcclusion::Template::Loader::Loader(Template &t):
	DataFile::DerivedObjectLoader<Template, PostProcessor::Template::Loader>(t)
{
	add("darkness", &Template::darkness);
	add("edge_depth_threshold", &Template::edge_depth_threshold);
	add("occlusion_radius", &Template::occlusion_radius);
	add("samples", &Template::n_samples);
}

} // namespace GL
} // namespace Msp
