#include <msp/core/raii.h>
#include <msp/strings/format.h>
#include "output.h"

using namespace std;

namespace Msp {
namespace GL {
namespace SL {

Formatter::Formatter():
	stage(0),
	source_index(0),
	source_line(1),
	indent(0),
	parameter_list(false),
	omit_builtin(false),
	r_empty_name(false)
{ }

string Formatter::apply(Stage &s)
{
	stage = &s;
	omit_builtin = true;

	const Version &ver = s.required_features.glsl_version;

	if(ver)
	{
		append(format("#version %d%02d", ver.major, ver.minor));
		if(s.required_features.gl_api==OPENGL_ES2 && ver>=Version(3, 0))
			append(" es");
		formatted += '\n';
	}

	if(s.required_features.arb_enhanced_layouts)
		append("#extension GL_ARB_enhanced_layouts: require\n");
	if(s.required_features.arb_explicit_attrib_location)
		append("#extension GL_ARB_explicit_attrib_location: require\n");
	if(s.required_features.arb_explicit_uniform_location)
		append("#extension GL_ARB_explicit_uniform_location: require\n");
	if(s.required_features.arb_gpu_shader5)
		append("#extension GL_ARB_gpu_shader5: require\n");
	if(s.required_features.arb_separate_shader_objects)
		append("#extension GL_ARB_separate_shader_objects: require\n");
	if(s.required_features.arb_uniform_buffer_object)
		append("#extension GL_ARB_uniform_buffer_object: require\n");
	if(s.required_features.ext_gpu_shader4)
		append("#extension GL_EXT_gpu_shader4: require\n");
	if(s.required_features.ext_texture_array)
		append("#extension GL_EXT_texture_array: require\n");
	formatted += '\n';

	s.content.visit(*this);

	return formatted;
}

void Formatter::append(const string &text)
{
	formatted += text;
	for(string::const_iterator i=text.begin(); i!=text.end(); ++i)
		if(*i=='\n')
			++source_line;
}

void Formatter::append(char c)
{
	formatted += c;
	if(c=='\n')
		++source_line;
}

void Formatter::set_source(unsigned index, unsigned line)
{
	if(index!=source_index || (index && line!=source_line))
	{
		if(index==source_index && line==source_line+1)
			formatted += '\n';
		else
		{
			unsigned l = line;
			if(stage && stage->required_features.glsl_version && stage->required_features.glsl_version<Version(3, 30))
				--l;
			formatted += format("#line %d %d\n", l, index);
		}
	}
	source_index = index;
	source_line = line;
}

void Formatter::visit(Block &block)
{
	unsigned brace_indent = indent;
	bool use_braces = (block.use_braces || (indent && block.body.size()!=1));
	if(use_braces)
		append(format("%s{\n", string(brace_indent*2, ' ')));

	SetForScope<unsigned> set(indent, indent+(indent>0 || use_braces));
	string spaces(indent*2, ' ');
	bool first = true;
	for(NodeList<Statement>::iterator i=block.body.begin(); i!=block.body.end(); ++i)
	{
		if(omit_builtin && (*i)->source<=BUILTIN_SOURCE)
			continue;
		if(!first)
			append('\n');
		first = false;
		set_source((*i)->source, (*i)->line);
		append(spaces);
		(*i)->visit(*this);
	}

	if(use_braces)
		append(format("\n%s}", string(brace_indent*2, ' ')));
}

void Formatter::visit_expression(Expression &expr, const Operator *outer_oper, bool on_rhs)
{
	unsigned outer_precedence = (outer_oper ? outer_oper->precedence : 20);
	unsigned inner_precedence = (expr.oper ? expr.oper->precedence : 0);

	bool needs_parentheses = (inner_precedence>=outer_precedence);

	// Omit parentheses if the outer operator encloses this operand.
	if(outer_oper && outer_oper->type==Operator::BINARY && outer_oper->token2[0] && on_rhs)
		needs_parentheses = false;

	if(expr.oper)
	{
		/* Omit parentheses if the inner expression's operator sits between the
		expression and the outer operator. */
		bool oper_on_left = expr.oper->type==Operator::PREFIX;
		bool oper_on_right = expr.oper->type==Operator::POSTFIX || (expr.oper->type==Operator::BINARY && expr.oper->token2[0]);
		if(expr.oper && ((oper_on_left && on_rhs) || (oper_on_right && !on_rhs)))
			needs_parentheses = false;

		// Omit parentheses if the operator's natural grouping works out.
		if(expr.oper==outer_oper)
			needs_parentheses = (expr.oper->assoc!=Operator::ASSOCIATIVE && on_rhs!=(expr.oper->assoc==Operator::RIGHT_TO_LEFT));
	}

	if(needs_parentheses)
		append('(');
	expr.visit(*this);
	if(needs_parentheses)
		append(')');
}

void Formatter::visit(Literal &literal)
{
	append(literal.token);
}

void Formatter::visit(VariableReference &var)
{
	append(var.name);
	r_empty_name = false;
}

void Formatter::visit(InterfaceBlockReference &iface)
{
	r_empty_name = iface.declaration->instance_name.empty();
	if(!r_empty_name)
		append(iface.declaration->instance_name);
}

void Formatter::visit(MemberAccess &memacc)
{
	visit_expression(*memacc.left, memacc.oper, false);
	if(!r_empty_name)
		append('.');
	append(memacc.member);
	r_empty_name = false;
}

void Formatter::visit(Swizzle &swizzle)
{
	visit_expression(*swizzle.left, swizzle.oper, false);
	append(format(".%s", swizzle.component_group));
}

void Formatter::visit(UnaryExpression &unary)
{
	if(unary.oper->type==Operator::PREFIX)
		append(unary.oper->token);
	visit_expression(*unary.expression, unary.oper, unary.oper->type==Operator::PREFIX);
	if(unary.oper->type==Operator::POSTFIX)
		append(unary.oper->token);
}

void Formatter::visit(BinaryExpression &binary)
{
	visit_expression(*binary.left, binary.oper, false);
	append(binary.oper->token);
	visit_expression(*binary.right, binary.oper, true);
	if(binary.oper->token2[0])
		append(binary.oper->token2);
}

void Formatter::visit(Assignment &assign)
{
	visit_expression(*assign.left, assign.oper, false);
	append(format(" %s ", assign.oper->token));
	visit_expression(*assign.right, assign.oper, true);
}

void Formatter::visit(TernaryExpression &ternary)
{
	visit_expression(*ternary.condition, ternary.oper, false);
	append(ternary.oper->token);
	visit_expression(*ternary.true_expr, ternary.oper, false);
	if(ternary.oper->token2)
		append(ternary.oper->token2);
	visit_expression(*ternary.false_expr, ternary.oper, true);
}

void Formatter::visit(FunctionCall &call)
{
	append(format("%s(", call.name));
	for(NodeArray<Expression>::iterator i=call.arguments.begin(); i!=call.arguments.end(); ++i)
	{
		if(i!=call.arguments.begin())
			append(", ");
		(*i)->visit(*this);
	}
	append(')');
}

void Formatter::visit(ExpressionStatement &expr)
{
	expr.expression->visit(*this);
	append(';');
}

void Formatter::visit(Import &import)
{
	append(format("import %s;", import.module));
}

void Formatter::visit(Precision &prec)
{
	append(format("precision %s %s;", prec.precision, prec.type));
}

void Formatter::visit(Layout &layout)
{
	append("layout(");
	for(vector<Layout::Qualifier>::const_iterator i=layout.qualifiers.begin(); i!=layout.qualifiers.end(); ++i)
	{
		if(i!=layout.qualifiers.begin())
			append(", ");
		append(i->name);
		if(i->has_value)
			append(format("=%d", i->value));
	}
	append(')');
}

void Formatter::visit(InterfaceLayout &layout)
{
	layout.layout.visit(*this);
	append(format(" %s;", layout.interface));
}

void Formatter::visit(StructDeclaration &strct)
{
	append(format("struct %s\n", strct.name));
	strct.members.visit(*this);
	append(';');
}

void Formatter::visit(VariableDeclaration &var)
{
	if(var.layout)
	{
		var.layout->visit(*this);
		append(' ');
	}
	if(var.constant)
		append("const ");
	if(!var.interpolation.empty())
		append(format("%s ", var.interpolation));
	if(!var.sampling.empty())
		append(format("%s ", var.sampling));
	if(!var.interface.empty())
	{
		string interface = var.interface;
		if(stage && stage->required_features.glsl_version && stage->required_features.glsl_version<Version(1, 30))
		{
			if(stage->type==Stage::VERTEX && var.interface=="in")
				interface = "attribute";
			else if((stage->type==Stage::VERTEX && var.interface=="out") || (stage->type==Stage::FRAGMENT && var.interface=="in"))
				interface = "varying";
		}
		append(format("%s ", interface));
	}
	if(!var.precision.empty())
		append(format("%s ", var.precision));
	string type_name = var.type_declaration->name;
	if(var.array)
		type_name = type_name.substr(0, type_name.find('['));
	append(format("%s %s", type_name, var.name));
	if(var.array)
	{
		append('[');
		if(var.array_size)
			var.array_size->visit(*this);
		append(']');
	}
	if(var.init_expression)
	{
		append(" = ");
		var.init_expression->visit(*this);
	}
	if(!parameter_list)
		append(';');
}

void Formatter::visit(InterfaceBlock &iface)
{
	if(iface.layout)
	{
		iface.layout->visit(*this);
		append(' ');
	}
	append(format("%s %s\n", iface.interface, iface.block_name));
	if(iface.struct_declaration)
		iface.struct_declaration->members.visit(*this);
	if(!iface.instance_name.empty())
	{
		append(' ');
		append(iface.instance_name);
		if(iface.array)
			append("[]");
	}
	append(';');
}

void Formatter::visit(FunctionDeclaration &func)
{
	append(format("%s %s(", func.return_type_declaration->name, func.name));
	for(NodeArray<VariableDeclaration>::iterator i=func.parameters.begin(); i!=func.parameters.end(); ++i)
	{
		if(i!=func.parameters.begin())
			append(", ");
		SetFlag set(parameter_list);
		(*i)->visit(*this);
	}
	append(')');
	if(func.definition==&func)
	{
		append('\n');
		func.body.visit(*this);
	}
	else
		append(';');
}

void Formatter::visit(Conditional &cond)
{
	append("if(");
	cond.condition->visit(*this);
	append(")\n");

	cond.body.visit(*this);
	if(!cond.else_body.body.empty())
	{
		Conditional *else_cond = dynamic_cast<Conditional *>(cond.else_body.body.front().get());
		if(cond.else_body.body.size()==1 && else_cond)
		{
			append('\n');
			set_source(else_cond->source, else_cond->line);
			append(format("%selse ", string(indent*2, ' ')));
			else_cond->visit(*this);
		}
		else
		{
			append(format("\n%selse\n", string(indent*2, ' ')));
			cond.else_body.visit(*this);
		}
	}
}

void Formatter::visit(Iteration &iter)
{
	if(!iter.init_statement && iter.condition && !iter.loop_expression)
	{
		append("while(");
		iter.condition->visit(*this);
		append(')');
	}
	else
	{
		append("for(");
		if(iter.init_statement)
			iter.init_statement->visit(*this);
		else
			append(';');
		if(iter.condition)
		{
			append(' ');
			iter.condition->visit(*this);
		}
		append(';');
		if(iter.loop_expression)
		{
			append(' ');
			iter.loop_expression->visit(*this);
		}
		append(')');
	}

	if(iter.body.body.empty())
		append(" { }");
	else
	{
		append('\n');
		iter.body.visit(*this);
	}
}

void Formatter::visit(Passthrough &pass)
{
	append("passthrough");
	if(pass.subscript)
	{
		append('[');
		pass.subscript->visit(*this);
		append(']');
	}
	append(';');
}

void Formatter::visit(Return &ret)
{
	append("return");
	if(ret.expression)
	{
		append(' ');
		ret.expression->visit(*this);
	}
	append(';');
}

void Formatter::visit(Jump &jump)
{
	append(jump.keyword);
	append(';');
}

} // namespace SL
} // namespace GL
} // namespace Msp
